/*
 * Copyright (c) 2025 Michael Hope <michaelh@juju.nz>
 *
 * SPDX-License-Identifier: Apache-2.0
 */

#include <hal_ch32fun.h>
#include <zephyr/device.h>
#include <zephyr/drivers/dma.h>
#include <zephyr/drivers/pwm.h>
#include <zephyr/kernel.h>
#include <zephyr/sys/printk.h>

#include <array>

#include "ch32x00xhw.h"
#include "line.h"
#include "pwm.h"
#include "timing.h"

namespace ch81 {
namespace {

constexpr uint8_t kHalt = 118;
constexpr uint8_t kSpace = 0;
constexpr uint8_t kDiagonal = 6;
constexpr uint8_t kShade = 8;
constexpr uint8_t kLeftBar = 5;

const uint8_t dfile[] = {
#if 0
    // Line 0
    kHalt, kDiagonal, kSpace, kShade, kHalt,
    // Line 1
    kSpace, kShade, kHalt,
#endif
    // Line 2
    0x2D, 0x2A, 0x31, 0x31, 0x34, 0x0D, kHalt,
    // Line 3
    kSpace, 0x2D, 0x2A, 0x31, 0x31, 0x34, 0x0D, kHalt,
    // Line 4
    kSpace, kSpace, 0x2D, 0x2A, 0x31, 0x31, 0x34, 0x0D, kHalt,
    // Line 5
    0x1C, 0x1D, 0x1E, 0x1F, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x1C, 0x1D,
    0x1E, 0x1F, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x1C, 0x1D, 0x1E, 0x1F,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x1C, 0x1D, kHalt,
    // Line 2
    0x2D, 0x2A, 0x31, 0x31, 0x34, 0x0D, kHalt,
    // Line 3
    kSpace, 0x2D, 0x2A, 0x31, 0x31, 0x34, 0x0D, kHalt,
    // Line 4
    kSpace, kSpace, 0x2D, 0x2A, 0x31, 0x31, 0x34, 0x0D, kHalt,
    // Line 5
    0x1C, 0x1D, 0x1E, 0x1F, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x1C, 0x1D,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x1C, 0x1D, kHalt,
    // Line 2
    0x2D, 0x2A, 0x31, 0x31, 0x34, 0x0D, kHalt,
    // Line 3
    kSpace, 0x2D, 0x2A, 0x31, 0x31, 0x34, 0x0D, kHalt,
    // Line 4
    kSpace, kSpace, 0x2D, 0x2A, 0x31, 0x31, 0x34, 0x0D, kHalt,
    // Line 5
    0x1C, 0x1D, 0x1E, 0x1F, 0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x1C, 0x1D,
    0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x1C, 0x1D, kHalt,
    // Remainder
    kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt,
    kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt,
    kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt,
    kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt, kHalt,
    kHalt, kHalt, kHalt, kHalt};

class DoubleBuffer {
 public:
  uint16_t* Next() { return buffers_[phase_ ^= 1].data(); }

 private:
  using Buffer = std::array<uint16_t, kTextWidth>;
  std::array<Buffer, 2> buffers_;
  //  uint16_t buffers_[2][kTextWidth * 2];
  int phase_ = 0;
};

static uint16_t to_write;

void QueueNextLine(SPI_TypeDef* spi, DMA_TypeDef* dma, uint16_t* line,
                   int wrote) {
  if ((DMA1_Channel6->CFGR & DMA_CFGR1_EN) != 0) {
    while ((dma->INTFR & DMA_TCIF6) == 0) {
    }
  }
  if ((DMA1_Channel3->CFGR & DMA_CFGR1_EN) != 0) {
    while ((dma->INTFR & DMA_TCIF3) == 0) {
    }
  }

  // Channel 3 is SPI_TX and TIM1_CH2.
  dma->INTFCR = DMA_CTCIF3 | DMA_CTCIF6;

  DMA1_Channel3->MADDR = reinterpret_cast<uintptr_t>(line);
  DMA1_Channel3->CFGR |= DMA_CFGR1_EN;

  to_write = wrote;
  DMA1_Channel6->CNTR = 1;
  DMA1_Channel6->CFGR |= DMA_CFGR1_EN;
}

[[nodiscard]]
int SetupHsync(uint64_t& cycles_per_sec) {
  TIM1->DMAINTENR |= TIM1_DMAINTENR_CC3DE;

  constexpr struct pwm_dt_spec hsync_dev = PWM_DT_SPEC_GET(DT_NODELABEL(hsync));
  if (int err = pwm_get_cycles_per_sec(hsync_dev.dev, hsync_dev.channel,
                                       &cycles_per_sec);
      err != 0) {
    return err;
  }

  if (int err = pwm_set(hsync_dev.dev, hsync_dev.channel, kLineNanos, 0,
                        hsync_dev.flags);
      err != 0) {
    return err;
  }

  constexpr struct pwm_dt_spec line_start_dev =
      PWM_DT_SPEC_GET(DT_NODELABEL(line_start));
  if (int err = pwm_set(line_start_dev.dev, line_start_dev.channel, kLineNanos,
                        kHsyncNanos + kBackPorchNanos, line_start_dev.flags);
      err != 0) {
    return err;
  }

  return 0;
}

[[nodiscard]]
int SetupDotDma(SPI_TypeDef* spi) {
  const struct device* dma_device =
      DEVICE_DT_GET(DT_DMAS_CTLR_BY_NAME(DT_NODELABEL(spi1), tx));
  uint8_t dot_channel = DT_DMAS_CELL_BY_NAME(DT_NODELABEL(spi1), tx, channel);

  dma_block_config dma_block = {
      .source_address = reinterpret_cast<uintptr_t>(&kHalt),
      .dest_address = reinterpret_cast<uintptr_t>(&spi->DATAR),
      .block_size = 0,
      .source_addr_adj = DMA_ADDR_ADJ_INCREMENT,
      .dest_addr_adj = DMA_ADDR_ADJ_NO_CHANGE,
  };

  struct dma_config config = {
      .channel_direction = MEMORY_TO_PERIPHERAL,
      .error_callback_dis = true,
      .channel_priority = 3,
      .source_data_size = 16,
      .dest_data_size = 16,
      .block_count = 1,
      .head_block = &dma_block,
  };

  if (int err = dma_config(dma_device, dot_channel, &config); err != 0) {
    return err;
  }

  return 0;
}

int SetupLineStartDma() {
  const struct device* dma_device = DEVICE_DT_GET(DT_NODELABEL(dma1));
  uint8_t line_start_channel = 5;

  dma_block_config dma_block = {
      .source_address = reinterpret_cast<uintptr_t>(&to_write),
      .dest_address = reinterpret_cast<uintptr_t>(&DMA1_Channel3->CNTR),
      .block_size = 0,
      .source_addr_adj = DMA_ADDR_ADJ_NO_CHANGE,
      .dest_addr_adj = DMA_ADDR_ADJ_NO_CHANGE,
  };

  struct dma_config config = {
      .channel_direction = MEMORY_TO_PERIPHERAL,
      .error_callback_dis = true,
      .channel_priority = 3,
      .source_data_size = 16,
      .dest_data_size = 16,
      .block_count = 1,
      .head_block = &dma_block,
  };

  if (int err = dma_config(dma_device, line_start_channel, &config); err != 0) {
    return err;
  }

  return 0;
}

[[nodiscard]]
int SetupSpi(SPI_TypeDef* spi) {
  spi->CTLR1 = SPI_CTLR1_SSM | SPI_CTLR1_SSI | SPI_CTLR1_MSTR | SPI_CTLR1_CPHA |
               SPI_CTLR1_DFF;
  spi->CTLR2 = SPI_CTLR2_TXDMAEN;
  spi->HSCR = SPI_HSCR_HSRXEN;
  spi->CTLR1 |= (kSpiPrescaler - 2) * SPI_CTLR1_BR_0;
  spi->STATR = 0;
  spi->CTLR1 |= SPI_CTLR1_SPE;

  return 0;
}

static DoubleBuffer buffers;

[[nodiscard]]
int Run() {
  arch_irq_lock();
  DevicetreePwmChannel<PWM_DT_SPEC_GET(DT_NODELABEL(hsync)),
                       DT_REG_ADDR(DT_NODELABEL(tim1))>
      hsync;
  DevicetreePwmChannel<PWM_DT_SPEC_GET(DT_NODELABEL(line_start)),
                       DT_REG_ADDR(DT_NODELABEL(tim1))>
      line_start;
  uint64_t cycles_per_sec;

  if (int err = SetupHsync(cycles_per_sec); err != 0) {
    return err;
  }

  SPI_TypeDef* spi =
      reinterpret_cast<SPI_TypeDef*>(DT_REG_ADDR(DT_NODELABEL(spi1)));
  if (int err = SetupSpi(spi); err != 0) {
    return err;
  }

  if (int err = SetupDotDma(spi); err != 0) {
    return err;
  }

  if (int err = SetupLineStartDma(); err != 0) {
    return err;
  }

  DMA_TypeDef* dma = reinterpret_cast<DMA_TypeDef*>(
      DT_REG_ADDR(DT_DMAS_CTLR_BY_NAME(DT_NODELABEL(spi1), tx)));

  // hsync.AwaitTop();

  for (;;) {
    for (int line = 0; line < kPreEqualizingLines; ++line) {
      hsync.AwaitTop();
      hsync.SetPulse(kHsyncNanos / kNanosPerPulse);
    }
    for (int line = 0; line < kVsyncLines; ++line) {
      hsync.AwaitTop();
      hsync.SetPulse((kLineNanos - kHsyncNanos - kBackPorchNanos) /
                     kNanosPerPulse);
    }
    for (int line = 0; line < kPostEqualizingLines; ++line) {
      hsync.AwaitTop();
      hsync.SetPulse(kHsyncNanos / kNanosPerPulse);
    }
    for (int line = 0; line < kPreBorderLines; ++line) {
      hsync.AwaitTop();
    }
    uint16_t* buffer = buffers.Next();
    int wrote;
    Render(&kHalt, 0, buffer, &wrote);
    for (int line = 0; line < kTopBorderLines; ++line) {
      QueueNextLine(spi, dma, buffer, wrote);
    }
    // Text.
    const uint8_t* ch = dfile;
    for (int i = 0; i < kTextLines; ++i) {
      uint16_t* buffer = buffers.Next();
      int wrote;
      const uint8_t* ch2 = Render(ch, i & 7, buffer, &wrote);
      if ((i & 7) == 7) {
        ch = ch2;
      }
      QueueNextLine(spi, dma, buffer, wrote);
    }
    buffer = buffers.Next();
    Render(&kHalt, 0, buffer, &wrote);
    for (int line = 0; line < kBottomBorderLines; ++line) {
      QueueNextLine(spi, dma, buffer, wrote);
    }
  }
}

}  // namespace
}  // namespace ch81

int main() {
  int err = ch81::Run();
  if (err != 0) {
    printk("Run() failed: %d\n", err);
  }
  return err;
}
